#include <vdr/ringbuffer.h>

#include "server/chanstreamer.h"
#include "remux/ts2ps.h"
#include "remux/ts2es.h"

cStreamdevChanReceiver::cStreamdevChanReceiver(const cChannel *Channel, 
		int Priority, cStreamdevChanStreamer *Streamer):
		cReceiver(Channel->Ca(), Priority, 7, Channel->Vpid(), Channel->Ppid(),
				Channel->Apid1(), Channel->Apid2(), Channel->Dpid1(), Channel->Dpid2(), 
				Channel->Tpid()) {
	m_Channel  = Channel;
	m_Streamer = Streamer;
}

cStreamdevChanReceiver::~cStreamdevChanReceiver() {
	Dprintf("Killing Channel Receiver\n");
	Detach();
}

void cStreamdevChanReceiver::Activate(bool On) {
	Dprintf("receiver activation %d\n", On);
	m_Streamer->Activate(On);
}

void cStreamdevChanReceiver::Receive(uchar *Data, int Length) {
	static time_t firsterr = 0;
	static int errcnt = 0;
	static bool showerr = true;

	int p = m_Streamer->Put(Data, Length);
	if (p != Length) {
		++errcnt;
		if (showerr) {
			if (firsterr == 0)
				firsterr = time_ms();
			else if (firsterr + BUFOVERTIME > time_ms() && errcnt > BUFOVERCOUNT) {
				esyslog("ERROR: too many buffer overflows, logging stopped");
				showerr = false;
				firsterr = time_ms();
			}
		} else if (firsterr + BUFOVERTIME < time_ms()) {
			showerr = true;
			firsterr = 0;
			errcnt = 0;
		}

		if (showerr)
			esyslog("ERROR: ring buffer overflow (%d bytes dropped)", Length - p);
		else
			firsterr = time_ms();
	}
}

cStreamdevChanStreamer::cStreamdevChanStreamer(int Priority, 
		eStreamType StreamType, cTBSocket *Socket):
		cStreamdevStreamer("Channel streamer", Socket) {
	m_Device     = NULL;
	m_Receiver   = NULL;
	m_Remux      = NULL;
	m_Priority   = Priority;
	m_StreamType = StreamType;
}

cStreamdevChanStreamer::~cStreamdevChanStreamer() {
	Dprintf("Killing Channel Streamer, Remuxer and Receiver\n");
	Stop(); // Must stop the thread before killing the Remuxer
	if (m_Remux    != NULL) delete m_Remux; 
	if (m_Receiver != NULL) delete m_Receiver;
}

void cStreamdevChanStreamer::DetachAll(void) {
	if (m_Receiver != NULL) m_Device->Detach(m_Receiver);
}

void cStreamdevChanStreamer::AttachAll(void) {
	if (m_Receiver != NULL) m_Device->AttachReceiver(m_Receiver);
}

bool cStreamdevChanStreamer::SetChannel(const cChannel *Channel) {
	m_Receiver = new cStreamdevChanReceiver(Channel, m_Priority, this);

	Dprintf("attaching to %p\n", m_Device);

	if (m_Device == NULL || !m_Device->AttachReceiver(m_Receiver))
		return false;

	Dprintf("now starting remuxer\n");

	switch (m_StreamType) {
	case stES:
		// XXX Send Audio Pid 1, otherwise Video Elementary Stream
		if (Channel->Vpid() == 0 || Channel->Vpid() == 1 
				|| Channel->Vpid() == 0x1FFF)
			m_Remux = new cTS2ESRemux(Channel->Apid1());
		else
			m_Remux = new cTS2ESRemux(Channel->Vpid());
		break;

	case stPES:
	case stPS:
		m_Remux = new cTS2PSRemux(Channel->Vpid(), Channel->Apid1(), 0, 0, 0, 
				m_StreamType == stPS);
		break;

	default:
		break;
	}
	return true;
}

uchar *cStreamdevChanStreamer::Process(const uchar *Data, int &Count, 
		int &Result) {
	if (m_Remux != NULL)
		return m_Remux->Process(Data, Count, Result);

	return cStreamdevStreamer::Process(Data, Count, Result);
}
