#include <vdr/ringbuffer.h>

#include "server/pidstreamer.h"
#include "common.h"

cStreamdevPidReceiver::cStreamdevPidReceiver(int Pid, int Priority, 
		cStreamdevPidStreamer *Streamer):
		cReceiver(0, Priority, 1, Pid) {
	m_Pid      = Pid;
	m_Streamer = Streamer;
}

cStreamdevPidReceiver::~cStreamdevPidReceiver() {
	Dprintf("Killing Pid Receiver for pid %d\n", m_Pid);
	Detach();
}

void cStreamdevPidReceiver::Activate(bool On) {
	Dprintf("receiver activation %d\n", On);
	m_Streamer->Activate(On);
}

void cStreamdevPidReceiver::Receive(uchar *Data, int Length) {
	static time_t firsterr = 0;
	static int errcnt = 0;
	static bool showerr = true;

	int p = m_Streamer->Put(Data, Length);
	if (p != Length) {
		++errcnt;
		if (showerr) {
			if (firsterr == 0)
				firsterr = time_ms();
			else if (firsterr + BUFOVERTIME > time_ms() && errcnt > BUFOVERCOUNT) {
				esyslog("ERROR: too many buffer overflows, logging stopped");
				showerr = false;
				firsterr = time_ms();
			}
		} else if (firsterr + BUFOVERTIME < time_ms()) {
			showerr = true;
			firsterr = 0;
			errcnt = 0;
		}

		if (showerr)
			esyslog("ERROR: ring buffer overflow (%d bytes dropped)", Length - p);
		else
			firsterr = time_ms();
	}
}

cStreamdevPidStreamer::cStreamdevPidStreamer(int Priority, cTBSocket *Socket):
		cStreamdevStreamer("Pid streamer", Socket) {
	m_NumReceivers = 0;
	m_Device       = NULL;
	m_Priority     = Priority;
}

cStreamdevPidStreamer::~cStreamdevPidStreamer() {
	Dprintf("Killing Pid Streamer and all receivers\n");
	for (int idx = 0; idx < m_NumReceivers; ++idx)
		delete m_Receivers[idx];
}

void cStreamdevPidStreamer::DetachAll(void) {
	for (int idx = 0; idx < m_NumReceivers; ++idx)
		m_Device->Detach(m_Receivers[idx]);
}

void cStreamdevPidStreamer::AttachAll(void) {
	for (int idx = 0; idx < m_NumReceivers; ++idx)
		m_Device->AttachReceiver(m_Receivers[idx]);
}

bool cStreamdevPidStreamer::SetPid(int Pid, bool On) {
	int idx;

	if (On) {
		if (m_NumReceivers == MAXNUMRECEIVERS) {
			esyslog("ERROR: Streamdev: No free receiver slots for pid %d", Pid);
			return false;
		}

		idx = m_NumReceivers++;
		m_Receivers[idx] = new cStreamdevPidReceiver(Pid, m_Priority, this);
		m_Device->AttachReceiver(m_Receivers[idx]);
	} else {
		for (idx = 0; idx < m_NumReceivers; ++idx)
			if (m_Receivers[idx]->m_Pid == Pid) break;

		if (idx == m_NumReceivers) {
			esyslog("ERROR: Streamdev: Trying to detach non-existing receiver");
			return false;
		}

		delete m_Receivers[idx];
		if (idx < --m_NumReceivers)
			memmove(&m_Receivers[idx], &m_Receivers[idx + 1], 
					(m_NumReceivers - idx) * sizeof(cStreamdevPidReceiver*));
	}
	return true;
}
