#include "server/server.h"
#include "server/componentVTP.h"
#include "server/componentHTTP.h"
#include "server/setup.h"

#include <vdr/tools.h>
#include <tools/select.h>
#include <string.h>
#include <errno.h>

cSVDRPhosts StreamdevHosts;

cStreamdevServer *cStreamdevServer::m_Instance = NULL;

cStreamdevServer::cStreamdevServer(void)
#if VDRVERSNUM >= 10300
		: cThread("Streamdev: server")
#endif
{
	m_Active = false;

	StreamdevHosts.Load(AddDirectory(cPlugin::ConfigDirectory(), 
			"streamdevhosts.conf"), true);
}

cStreamdevServer::~cStreamdevServer() {
	if (m_Active) Stop();
}

void cStreamdevServer::Init(void) {
	if (m_Instance == NULL) {
		m_Instance = new cStreamdevServer;
		if (StreamdevServerSetup.StartVTPServer)
			m_Instance->Register(new cComponentVTP);
		if (StreamdevServerSetup.StartHTTPServer)
			m_Instance->Register(new cComponentHTTP);
		m_Instance->Start();
	}
}

void cStreamdevServer::Exit(void) {
	if (m_Instance != NULL) {
		m_Instance->Stop();
		DELETENULL(m_Instance);
	}
}

void cStreamdevServer::Stop(void) {
	m_Active = false;
	Cancel(3);
}

void cStreamdevServer::Register(cServerComponent *Server) {
	m_Servers.Append(Server);
}

void cStreamdevServer::Action(void) {
	cTBSelect select;
	uint it;

#if VDRVERSNUM < 10300
	isyslog("Streamdev: Server thread started (pid=%d)", getpid());
#endif

	m_Active = true;

	/* Initialize Server components, deleting those that failed */
	for (it = 0; it < m_Servers.Count(); ++it) {
		if (!m_Servers[it]->Init()) 
			delete m_Servers.Take(it--);
	}

	if (!m_Servers.Count()) {
		esyslog("Streamdev: No server components registered, exiting");
		m_Active = false;
	}

	while (m_Active) {
		select.Clear();

		/* Ask all Server components to register to the selector */
		for (it = 0; it < m_Servers.Count(); ++it)
			m_Servers[it]->AddSelect(select);
		
		/* Ask all Client connections to register to the selector */
		for (it = 0; it < m_Clients.Count(); ++it) 
			m_Clients[it]->AddSelect(select);

		if (select.Select(1000) < 0) {
			if (!m_Active) // Exit was requested while polling
				continue;
			esyslog("Streamdev: Fatal error, server exiting: %s", strerror(errno));
			m_Active = false;
		}
	
		/* Ask all Server components to act on signalled sockets */
		for (it = 0; it < m_Servers.Count(); ++it) {
			cServerConnection *client;
			if ((client = m_Servers[it]->CanAct(select)) != NULL) {
				m_Clients.Append(client);

				if (m_Clients.Count() > (uint)StreamdevServerSetup.MaxClients) {
					esyslog("Streamdev: Too many clients, rejecting %s:%d",
							(const char*)client->RemoteIp(), client->RemotePort());
					client->Reject();
				} else if (!StreamdevHosts.Acceptable(client->RemoteIpAddr())) {
					esyslog("Streamdev: Client from %s:%d not allowed to connect",
							(const char*)client->RemoteIp(), client->RemotePort());
					client->Reject();
				} else 
					client->Welcome();
			}
		}

		/* Ask all Client connections to act on signalled sockets */
		for (it = 0; it < m_Clients.Count(); ++it) {
			if (!m_Clients[it]->CanAct(select)) {
				isyslog("Streamdev: Closing connection to %s:%d", 
						(const char*)m_Clients[it]->RemoteIp(), 
						m_Clients[it]->RemotePort());
				m_Clients[it]->Close();
				delete m_Clients.Take(it--);
			}
		}
	}
	
	while (m_Clients.Count()) {
		cServerConnection *client = m_Clients.TakeFirst();
		client->Close();
		delete client;
	}

	while (m_Servers.Count()) {
		cServerComponent *server = m_Servers.TakeFirst();
		server->Exit();
		delete server;
	}

#if VDRVERSNUM < 10300
	isyslog("Streamdev: Server thread stopped");
#endif
}
